# sparksdr2wsjtx-udp.py (2024-05-01) by AndyZ K1RA https://k1ra.us/
#
# Usage: sparksdr2wsjtx-udp.py -m MYCALL -g MYGRID6 [-c cache_expire] [-s sparksdr_ip] [-b broadcast_ip] [-p
# broadcast_port] [-d] [-v] [-h]
#
# A WSJT-X UDP broadcaster providing spots from SparkSDR https://sparksdr.com/ Websocket data
# for FT8, FT4, WSPR, FST4W, JS8 decodes across multiple bands to a multicast or direct UDP port
#
# sends WSJT-X UDP packets per definition
# https://sourceforge.net/p/wsjt/wsjtx/ci/master/tree/Network/NetworkMessage.hpp
#
# use "pip install -r requirements.txt" to install python dependencies and libraries needed by this app

import socket
import struct
import binascii
import random

import json
import os
import re
import time
import threading
import sys
import getopt
import psutil as psutil

from multiprocessing import Process, Queue
from websocket import WebSocketApp
from threading import Thread
from queue import Queue
from functools import partial

# must report ID as WSJT-X, else GridTracker won't accept packets
ID = 'WSJT-X'

revision = "SparkSDR2WSJTX-UDP"
version = "2.0.8"

# WSJT-X UDP header
header = binascii.unhexlify("adbccbda00000002")

# Message descriptors
# Heartbeat
msg0_hb = binascii.unhexlify("00000000")

# Status
msg1_status = binascii.unhexlify("00000001")

# Decode FTn
msg2_ftn = binascii.unhexlify("00000002")

# Decode WSPR
msg10_wspr = binascii.unhexlify("0000000A")

# Close
msg6_close = binascii.unhexlify("00000006")

# QT5 UDP schema
maxschema = binascii.unhexlify("00000003")

# variables used in UDP spot packets see https://sourceforge.net/p/wsjt/wsjtx/ci/master/tree/Network/NetworkMessage.hpp
txen = 0
tx = 0
decon = 1
decoff = 0
rxdf = 1024
txdf = 1024
txwat = 0
submode = ""
fast = 0


# Create a Timer class for scheduling tasks (currently not used)
class RepeatedTimer(object):
    def __init__(self, interval, function, *args, **kwargs):
        self._timer = None
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.is_running = False
        self.next_call = time.time()
        self.start()

    def _run(self):
        self.is_running = False
        self.start()
        self.function(*self.args, **self.kwargs)

    def start(self):
        if not self.is_running:
            self.next_call += self.interval
            self._timer = threading.Timer(self.next_call - time.time(), self._run)
            self._timer.start()
            self.is_running = True

    def stop(self):
        self._timer.cancel()
        self.is_running = False


# Create a Websocket client class for connecting to SparkSDR server
class Websocket_Client:
    def __init__(self, wsurl: str):
        self.responseQ = Queue()  # our message response Q

        # add the response Qs on by wrapping the callbacks in functools partials
        self.websocket = WebSocketApp(wsurl,
                                      on_open=partial(self.on_open, responseQ=self.responseQ),
                                      on_message=partial(self.on_message, responseQ=self.responseQ),
                                      on_error=partial(self.on_error, responseQ=self.responseQ),
                                      on_close=partial(self.on_close, responseQ=self.responseQ))

        # start run_forever as thread. request keep alive pings every 10 seconds with 5 second timeouts
        self.controller = Thread(target=self.websocket.run_forever, args=(None, None, 10, 5))
        self.controller.start()

    # open connection and send subscribe message to begin receiving spots
    def on_open(self, websocket, responseQ):
        global debug

        if debug:
            print("Opening Connection")
        time.sleep(1)
        websocket.send("{\"cmd\":\"subscribeToSpots\",\"Enable\":true}\n\n")
        time.sleep(1)
        websocket.send("{\"cmd\":\"subscribeToSpots\",\"Enable\":true}\n\n")

    def on_error(self, websocket, error, responseQ):
        global debug

        if debug:
            print("Connection Error:", error)

    # get a SparkSDR spot decode message and copy to the queue for processing
    def on_message(self, websocket, message, responseQ):
        global debug

        if debug:
            print("Message Received:", message)
        responseQ.put(message)

    # connection closed and stop receiving sport, but spawn reconnect attempts every 2 seconds
    def on_close(self, websocket, status, message, responseQ):
        global debug

        if debug:
            print("Closing Connection:")
        time.sleep(2)

        self.controller = Thread(target=self.websocket.run_forever, args=(None, None, 10, 5))
        self.controller.start()

    # force a connection close
    def stop(self):
        self.websocket.keep_running = False
        self.websocket.close()


# Parse a SparkSDR Websocket JSON message into individual fields
def parse_message(json_message):
    global debug
    global calls_db
    global cache_expire

    date = json_message["time"]

    # parse time
    m = re.search('(\d{2}):(\d{2}):(\d{2})', date)
    hh = m.group(1)
    mm = m.group(2)
    ss = m.group(3)
    hhmmss = f"{hh}{mm}{ss}"

    call = json_message["call"]
    mode = json_message["mode"]

    freq = json_message["tunedfrequency"]
    freq = freq / 1000.

    tone = int(json_message["frequency"]) - int(json_message["tunedfrequency"])
    dist = json_message["distance"]
    snr = str(json_message["snr"])
    loc = json_message["locator"]

    # parse grid
    if loc == "":
        grid = "    "
    else:
        m = re.search('([A-Z][A-Z]\d\d)', loc)
        grid = m.group(1)

    dt = float(json_message["dt"])
    # valid = json_message["valid"]
    valid = ""
    jtmsg = json_message["msg"]

    jtmsg = re.sub('[\r\n]', '', jtmsg)

    called = ""

    # handle FTn messages
    if re.search('FT\d', mode) is not None:
        # look for different FTn message structures and extract calls and grid

        # CQ CALL
        if (m := re.search('^CQ\s([A-Z\d\/]{3,})$', jtmsg)) is not None:
            call = m.group(1)

        # CQ CALL LLnn
        elif (m := re.search('^CQ\s([A-Z\d\/]{3,})\s([A-Z][A-Z]\d\d$)', jtmsg)) is not None:
            call = m.group(1)
            grid = m.group(2)

        # [NA, DX, SD, TEST, xxx, etc] CALL LLnn
        elif (m := re.search('^[A-Z]{2,}\s([A-Z\d\/]{3,})\s([A-Z][A-Z]\d\d$)', jtmsg)) is not None:
            call = m.group(1)
            grid = m.group(2)

        # CQ [NA, DX, VT, TEST, xxx, etc] CALL
        elif (m := re.search('^CQ\s[A-Z]{2,}\s([A-Z\d\/]{3,})$', jtmsg)) is not None:
            call = m.group(1)
            grid = ""

        # CQ [NA, DX, VT, TEST, xxx, etc] CALL LLnn
        elif (m := re.search('^CQ\s[A-Z]{2,}\s([A-Z\d\/]{3,})\s([A-Z][A-Z]\d\d$)', jtmsg)) is not None:
            call = m.group(1)
            grid = m.group(2)

        # CALL1 CALL2
        elif (m := re.search('^\<?(\.\.\.|[A-Z\d\/]{3,})\>?\s\<?([A-Z\d\/]{3,})\>?$', jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = ""

        # CALL1 CALL2 [R][-+]nn
        elif (m := re.search('^\<?(\.\.\.|[A-Z\d\/]{3,})\>?\s\<?([A-Z\d\/]{3,})\>?\sR*[\-+][0-9]{2}$', jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = ""

        # CALL1 CALL2 RRR
        elif (m := re.search('^\<?(\.\.\.|[A-Z\d\/]{3,})\>?\s\<?([A-Z\d\/]{3,})\>?\sRRR$', jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = ""

        # CALL1 CALL2 GRID
        elif (m := re.search('^\<?(\.\.\.|[A-Z\d\/]{3,})\>?\s\<?([A-Z\d\/]{3,})\>?\s([A-Z][A-Z]\d\d$)', jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = m.group(3)

        # CALL GRID
        elif (m := re.search('^\s*([A-Z\d\/]{3,})\s([A-Z][A-Z]\d\d)$', jtmsg)) is not None:
            call = m.group(1)
            grid = m.group(2)

        # CALL1 CALL2 RR73 or 73
        elif (m := re.search('^\<?(\.\.\.|[A-Z\d\/]{3,})\>?\s\<?([A-Z\d\/]{3,})\>?\sR*73$', jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = ""

        # SP9VEY RR73; IK6CLX <TY0RU> -08
        elif (m := re.search('^[A-Z\d\/]{3,}\sRR73;\s([A-Z\d\/]{3,})\s\<(\.\.\.|[A-Z\d\/]{3,})\>\s', jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = ""

        # CALL1 CALL2 R GRID
        elif (m := re.search('^\<?(\.\.\.|[A-Z\d\/]{3,})\>?\s([A-Z\d\/]{3,})\sR\s([A-Z][A-Z]\d\d$)', jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = m.group(3)

        # CALL1 CALL2 RST nnnn
        elif (m := re.search('^\<?(\.\.\.|[A-Z\d\/]{3,})\>?\s([A-Z\d\/]{3,})\s\d{3}\s\d{4}$', jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = ""
        # no match
        else:
            if debug:
                if call != "":
                    print(f"<<<--- NO MATCH --->>> {call}, {jtmsg}")
            return False, hhmmss, freq, mode, call, grid, tone, snr, dt, "", jtmsg

        if re.search('RR73', grid) is not None:
            grid = ""

    # pull fields only found in WSPR messages
    if mode == "WSPR" or mode == "FST4W":
        # WSPR
        # watt = json_message["watts"]
        dbm = json_message["power"]
        jtmsg = f"CQ {call} {grid}"

        if re.search('[A-Z][A-Z]\d\d', grid) is None:
            return False, hhmmss, freq, mode, call, grid, tone, snr, dt, dbm, jtmsg
    else:
        watt = ""
        dbm = ""

    call = re.sub('[<>\.]', '', call)

    # perform some sanity checks on calls, check for bogus combos and return False if necessary
    if (call == "" or re.search('RR73', call) is not None) or (re.search('\d[A-Z]', call) is None) or \
            (re.search('^\d{2;}[A-Z]{2,}', call) is not None) or (re.search('^[0Q]', call) is not None) or \
            (re.search('^[0Q]', call) is not None) or (re.search('^\d.\d.*\/[PQR]$', call) is not None) or \
            (re.search('^\d{2;}[A-Z]{2,}', called) is not None) or (re.search('_|\/.*\/]', call) is not None) or \
            (re.search('\d[A-Z]+\d[A-Z]+\d', call) is not None) or \
            ((re.search('\/[PQR]$', call) is not None) and (re.search('\/[PQR]$', called) is not None)) or \
            ((re.search('K1ABC', called) is not None) and (re.search('\/[PQR]$', call) is not None)):
        return False, hhmmss, freq, mode, call, grid, tone, snr, dt, dbm, jtmsg

    # create hash of the call sign and frequency spotted
    callfreq = f"{call}{freq}"

    # get current time
    nowtime = int(time.time())

    # test if call was seen in last N expire_cache minutes
    if callfreq in calls_db:
        # call has been seen
        spottime = calls_db[callfreq]

        # test if call passed expiration limit
        if spottime <= nowtime - (cache_expire * 60):
            # yes, reset expire time and return decode
            calls_db[callfreq] = nowtime
        else:
            # no, its fresh and recently seen
            valid = "FRESH"
    else:
        # call not seen yet, save in database and return decode
        calls_db[callfreq] = nowtime

    return valid, hhmmss, freq, mode, call, grid, tone, snr, dt, dbm, jtmsg


# Expire decoded calls in the database that haven't been seen in (cache_expire) minutes
def clean_calls_db():
    global debug
    global calls_db
    global cache_expire

    if debug:
        print("Cleaning cached calls database...")

    # get current time in seconds
    nowtime = int(time.time())

    db_len = len(calls_db)

    # iterate through all call+freq hashes in database and drop any older than cache_expire timer (minutes)
    db = {callfreq: s_time for callfreq, s_time in calls_db.items() if s_time > nowtime - (cache_expire * 60)}

    if debug:
        calls_db = db
        new_len = len(db)
        print(f"{db_len - new_len} items removed from database")
        print()


# Create WSJT-X heartbeat message
def hbeat_message():
    global debug
    global header, msg0_hb, ID, maxschema, version, revision

    s_id = str(len(ID)) + "s"
    s_ver = str(len(version)) + "s"
    s_rev = str(len(revision)) + "s"

    # A Msg 0 - Heartbeat
    pack = (struct.pack("8s", header) + struct.pack("4s", msg0_hb) + struct.pack(">I", len(ID)) +
            struct.pack(s_id, bytes(ID, "utf-8")) + struct.pack("4s", maxschema) + struct.pack(">I", len(version)) +
            struct.pack(s_ver, bytes(version, "utf-8")) + struct.pack(">I", len(revision)) +
            struct.pack(s_rev, bytes(revision, "utf-8")))

    return pack


# Create WSJT-X start message
def start_message(freq, mode, call, grid, snr, mycall, mygrid):
    global debug
    global header, msg1_status, ID, tx, decon, decoff, rxdf, txdf, txwat, submode, fast

    s_id = str(len(ID)) + "s"
    s_mode = str(len(mode)) + "s"
    s_call = str(len(call)) + "s"
    s_grid = str(len(grid)) + "s"
    s_snr = str(len(snr)) + "s"
    s_mycall = str(len(mycall)) + "s"
    s_mygrid = str(len(mygrid)) + "s"
    s_submode = str(len(submode)) + "s"

    # A Msg 1 - Location station info and decoding started
    pack = (struct.pack("8s", header) + struct.pack("4s", msg1_status) + struct.pack(">I", len(ID)) +
            struct.pack(s_id, bytes(ID, "utf-8")) + struct.pack(">Q", int(freq * 1000)) +
            struct.pack(">I", len(mode)) + struct.pack(s_mode, bytes(mode, "utf-8")) + struct.pack(">I", len(call)) +
            struct.pack(s_call, bytes(call, "utf-8")) + struct.pack(">I", len(snr)) +
            struct.pack(s_snr, bytes(snr, "utf-8")) + struct.pack(">I", len(mode)) +
            struct.pack(s_mode, bytes(mode, "utf-8")) + struct.pack("b", txen) + struct.pack("b", tx) +
            struct.pack("b", decon) + struct.pack(">I", rxdf) + struct.pack(">I", txdf) +
            struct.pack(">I", len(mycall)) + struct.pack(s_mycall, bytes(mycall, "utf-8")) + struct.pack(">I",
                                                                                                         len(mygrid)) +
            struct.pack(s_mygrid, bytes(mygrid, "utf-8")) + struct.pack(">I", len(grid)) +
            struct.pack(s_grid, bytes(grid, "utf-8)")) + struct.pack("b", txwat) + struct.pack(">I", len(submode)) +
            struct.pack(s_submode, bytes(submode, "utf-8")) + struct.pack("b", fast) + struct.pack("b", 0) +
            struct.pack(">I", 0) + struct.pack(">I", 0) + struct.pack(">I", len("Default")) +
            struct.pack("7s", b"Default") + struct.pack(">I", 0) + struct.pack(">I", 0))

    return pack


# Create WSJT-X decode message
def decode_message(hhmmss, call, grid, dt, tone, mode, snr, jtmsg, dbm):
    global debug
    global header, msg2_ftn, msg10_wspr, ID, version

    # parse time
    m = re.search('(\d{2})(\d{2})(\d{2})', hhmmss)
    hh = int(m.group(1))
    mm = int(m.group(2))
    ss = int(m.group(3))
    hhmmss = (hh * 3600 + mm * 60 + ss)

    s_id = str(len(ID)) + "s"
    s_call = str(len(call)) + "s"
    s_grid = str(len(grid)) + "s"
    s_mode = str(len(mode)) + "s"
    s_jtmsg = str(len(jtmsg)) + "s"

    if mode == "WSPR":
        # A Msg 10 - WSPR decode message
        pack = (struct.pack("8s", header) + struct.pack("4s", msg10_wspr) + struct.pack(">I", len(ID)) +
                struct.pack(s_id, bytes(ID, "utf-8")) + struct.pack("b", 1) +
                struct.pack(">I", hhmmss * 1000) + struct.pack(">i", int(snr)) +
                struct.pack(">d", dt) + struct.pack(">Q", tone) + struct.pack(">i", 0) +
                struct.pack(">I", len(call)) + struct.pack(s_call, bytes(call, "utf-8")) +
                struct.pack(">I", len(grid)) + struct.pack(s_grid, bytes(grid, "utf-8")) +
                struct.pack(">i", dbm) + struct.pack("b", 0))

    # A Msg 2 - FTn decode message
    pack = (struct.pack("8s", header) + struct.pack("4s", msg2_ftn) + struct.pack(">I", len(ID)) +
            struct.pack(s_id, bytes(ID, "utf-8")) + struct.pack("b", 1) +
            struct.pack(">I", hhmmss * 1000) + struct.pack(">i", int(snr)) +
            struct.pack(">d", dt) + struct.pack(">I", tone) + struct.pack(">I", len(mode)) +
            struct.pack(s_mode, bytes(mode, "utf-8")) + struct.pack(">I", len(jtmsg)) +
            struct.pack(s_jtmsg, bytes(jtmsg, "utf-8")) + struct.pack("b", 0) + struct.pack("b", 0))

    return pack


# Create WSJT-X stop message
def stop_message(freq, mode, call, grid, snr, mycall, mygrid):
    global debug
    global header, msg1_status, ID, tx, decon, decoff, rxdf, txdf, txwat, submode, fast

    s_id = str(len(ID)) + "s"
    s_mode = str(len(mode)) + "s"
    s_call = str(len(call)) + "s"
    s_grid = str(len(grid)) + "s"
    s_snr = str(len(snr)) + "s"
    s_mycall = str(len(mycall)) + "s"
    s_mygrid = str(len(mygrid)) + "s"
    s_submode = str(len(submode)) + "s"

    # A Msg 1 - Location station info and decoding started
    pack = (struct.pack("8s", header) + struct.pack("4s", msg1_status) + struct.pack(">I", len(ID)) +
            struct.pack(s_id, bytes(ID, "utf-8")) + struct.pack(">Q", int(freq * 1000)) +
            struct.pack(">I", len(mode)) + struct.pack(s_mode, bytes(mode, "utf-8")) + struct.pack(">I", len(call)) +
            struct.pack(s_call, bytes(call, "utf-8")) + struct.pack(">I", len(snr)) +
            struct.pack(s_snr, bytes(snr, "utf-8")) + struct.pack(">I", len(mode)) +
            struct.pack(s_mode, bytes(mode, "utf-8")) + struct.pack("b", txen) + struct.pack("b", tx) +
            struct.pack("b", decoff) + struct.pack(">I", rxdf) + struct.pack(">I", txdf) +
            struct.pack(">I", len(mycall)) + struct.pack(s_mycall, bytes(mycall, "utf-8")) + struct.pack(">I",
                                                                                                         len(mygrid)) +
            struct.pack(s_mygrid, bytes(mygrid, "utf-8")) + struct.pack(">I", len(grid)) +
            struct.pack(s_grid, bytes(grid, "utf-8)")) + struct.pack("b", txwat) + struct.pack(">I", len(submode)) +
            struct.pack(s_submode, bytes(submode, "utf-8")) + struct.pack("b", fast) + struct.pack("b", 0) +
            struct.pack(">I", 0) + struct.pack(">I", 0) + struct.pack(">I", len("Default")) +
            struct.pack("7s", b"Default") + struct.pack(">I", 0) + struct.pack(">I", 0))

    return pack


# socket for UDP connectio to GridTracker
sock = 0

# Start sparksdr2wsjtx-udp.py
if __name__ == '__main__':
    debug = False
    cache_expire = 5
    sparksdr_ip = "127.0.0.1"
    sparksdr_port = 4649
    broadcast_ip = "239.1.1.1"
    broadcast_port = 2237
    mycall = ""
    mygrid = ""

    try:
        opts, args = getopt.getopt(sys.argv[1:], "m:g:c:s:b:p:dvh")
    except getopt.GetoptError:
        print(
            "sparksdr2wsjtx-udp.py -m MYCALL -g MYGRID6 [-c cache_expire] [-s sparksdr_ip] [-b broadcast_ip] [-p "
            "broadcast_port] [-d] [-v] [-h]")
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print("Usage: sparksdr2wsjtx-udp.py -m MYCALL -g MYGRID6 [-c cache_expire] [-s sparksdr_ip] [-b "
                  "broadcast_ip] [-p broadcast_port] [-d] [-v] [-h]")
            print("-m MYCALL (mandatory, ex: -m K1RA)")
            print("-g MYGRID6 (mandatory, ex: -g FM18cr)")
            print("-c cache_expire (optional, minutes, ex: -c 5) [default=5]")
            print("-s sparksdr_ip (optional, ip address, ex: -s 192.168.1.123) [default=127.0.0.1]")
            print("-b broadcast_ip (optional, ip address, ex: -b 192.168.1.234) [default=239.1.1.1]")
            print("-p broadcast_port (optional, udp port, ex: -p 2239) [default=2237]")
            print("-d (turns on debug messages)")
            print("-v (prints version)")
            print("-h (prints this help)")
            sys.exit()
        elif opt in "-c":
            cache_expire = int(arg)
        elif opt in "-d":
            debug = True
        elif opt in "-m":
            mycall = arg
        elif opt in "-g":
            mygrid = arg
            if re.search('\w\w\d\d\w\w', mygrid) is None:
                print(f"6 digit grid {mygrid} incorrect")
                sys.exit(3)
        elif opt in "-s":
            sparksdr_ip = arg
            if re.search('\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}', sparksdr_ip) is None:
                print(f"sparksdr_ip address {sparksdr_ip} incorrect")
                sys.exit(3)
        elif opt in "-b":
            broadcast_ip = arg
            if re.search('\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}', broadcast_ip) is None:
                print(f"broadcast_ip address {broadcast_ip} incorrect")
                sys.exit(3)
        elif opt in "-p":
            port = arg
            broadcast_port = int(port)
            if broadcast_port < 1024 or broadcast_port > 65535:
                print(f"broadcast_port number {port} incorrect (1024-65535)")
                sys.exit(3)
        elif opt in "-v":
            print(f"sparksdr2wsjtx-udp version {version} by AndyZ K1RA")
            sys.exit()

    if mycall == "":
        print("MYCALL call sign missing.")
        print("sparksdr2wsjtx-udp.py -m MYCALL")
        sys.exit(2)

    if mygrid == "":
        print("MYGRID 6 digit grid missing.")
        print("sparksdr2wsjtx-udp.py -m MYCALL -g MYGRID6")
        sys.exit(2)

    multicast_group = (broadcast_ip, broadcast_port)

    print(f"sparksdr2wsjtx-udp version {version} by K1RA")

    url = f"ws://{sparksdr_ip}:{sparksdr_port}/Spark"

    # call+freq dictionary of received calls
    calls_db = {}

    # make the websocket client
    wsclient = Websocket_Client(url)

    pid = os.getpid()

    freq = 14074
    call = mycall
    grid = mygrid
    mode = "FT8"

    clean_timer = int(time.time())

    try:
        m = re.search('^(\d{1,3})\.\d{1,3}\.\d{1,3}\.\d{1,3}', broadcast_ip)
        hi_octet = int(m.group(1))
        # print(hi_octet)

        # check if we're broadcasting to localhost or a specific IP
        if hi_octet < 224:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

            port = random.randint(62700, 64000)

            while True:
                try:
                    sock.bind(('', port))
                    break
                except socket.error:
                    print(port, "UDP port in use, trying to find another send port...")
                    time.sleep(1)
                    port += 1
        else:
            # Create the datagram socket for Multicast
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

            port = random.randint(62700, 62700)

            while True:
                try:
                    sock.bind(('', port))
                    break
                except socket.error:
                    print(port, "Multicast port is use, trying to find another send port...")
                    time.sleep(1)
                    port += 1

            mreq = struct.pack("=4sl", socket.inet_aton(broadcast_ip), socket.INADDR_ANY)

            sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)

        message_hbeat = hbeat_message()
        sent = sock.sendto(message_hbeat, multicast_group)

        # do forever
        while True:
            # process FT8, FT4, WSPR websocket messages from SparkSDR queue
            while not wsclient.responseQ.empty():
                spark_msgs = wsclient.responseQ.get()

                json_msgs = json.loads(spark_msgs)

                printed = False

                # for every individual decode spot
                if "spots" in json_msgs:
                    for json_msg in json_msgs["spots"]:

                        valid, hhmmss, freq, mode, call, grid, tone, snr, dt, dbm, jtmsg = parse_message(json_msg)

                        # don't process or report call if it was recently seen within expire timeout
                        if valid == "FRESH" or call == "":
                            continue

                        if debug and (valid is False or valid == "false"):
                            print()
                            print("<<<--- INVALID --->>> {:6}  {:7.1f} {:>5} {:>11} {:>4} {:>3} {:>4}  {}".format(
                                hhmmss, freq, mode, call, tone, snr, grid, jtmsg))
                            print()
                            continue

                        # def start_message(freq, mode, call, grid, snr, mycall, mygrid):
                        message_start = start_message(freq, mode, call, grid, "0", mycall, mygrid)
                        sent = sock.sendto(message_start, multicast_group)

                        # Send every client the spot message
                        message_decode = decode_message(hhmmss, call, grid, dt, tone, mode, snr, jtmsg, dbm)
                        sent = sock.sendto(message_decode, multicast_group)

                        # def stop_message(freq, mode, call, grid, snr, mycall, mygrid):
                        message_stop = stop_message(freq, mode, call, grid, "0", mycall, mygrid)
                        sent = sock.sendto(message_stop, multicast_group)

                        if debug:
                            # print(hhmmss, freq, mode, call, tone, snr, grid, jtmsg, sep=" ")
                            print("{:6}  {:7.1f} {:>5} {:>11} {:>4} {:>3} {:>4}  {}".format(hhmmss, freq, mode,
                                                                                            call, tone, snr, grid,
                                                                                            jtmsg))
                            printed = True

                    if debug and printed:
                        print()

            # time.sleep(1)

            now_time = int(time.time())

            # once a minute run database expire clean and send heartbeat over network
            if now_time >= clean_timer + (1 * 60):
                clean_calls_db()
                clean_timer = int(time.time())

                message_hbeat = hbeat_message()
                sent = sock.sendto(message_hbeat, multicast_group)

    except KeyboardInterrupt:
        sock.close()

        if debug:
            print("Closing main thread")

        p = psutil.Process(pid)
        p.terminate()  # or p.kill()

        wsclient.controller.join()
